/*********************************************************

Programm zur Kommunikation zwischen MSP430F149 und MMC per SPI im unprotected Mode.
Sytemquarz: 8 MHz, Verdrahtung: siehe mmc.h.

Version 0.02 vom 11.5.2003

Stand: Lesen der Register funktioniert meistens, Lesen der Daten meist nur beim ersten Mal,
        Schreiben sehr selten, Initialisieren und setzen der Blocklaenge fast immer.

*********************************************************/
#include  <msp430x14x.h>
#include "mmc.h"

extern char card_state;

extern char mmc_buffer[512];

char card_state = 0;		// Zustand der Karte: 0 fuer nicht gefunden, 1 fuer gefunden


void
main (void)
{

  volatile unsigned char dco_cnt = 255;

  BCSCTL1 &= ~XT2OFF;		// XT2on
  do				// wait for MCLK from quartz
    {
      IFG1 &= ~OFIFG;		// Clear OSCFault flag
      for (dco_cnt = 0xff; dco_cnt > 0; dco_cnt--);	// Time for flag to set
    }
  while ((IFG1 & OFIFG) != 0);	// OSCFault flag still set?

  WDTCTL = WDTPW + WDTHOLD;	// Stop WDT

  BCSCTL1 = 0x07;		// LFXT1: XT2on: LF quartz for MCLK

  BCSCTL2 = SELM1 + SELS;	// LFXT2: use HF quartz (XT2)

  DCOCTL = 0xe0;

  P3SEL = 0x00;
  P3DIR = 0xcf;
  P3OUT = 0x0f;

  P4SEL = 0x7e;
  P4DIR = 0x60;
  P4OUT = 0x40;

  P5SEL = 0x00;
  P5DIR = 0xff;
  P5OUT = 0xfe;

  P6SEL = 0x00;
  P6OUT = 0x00;
  P6DIR = 0xff;

  ADC12IE = 0;
  ADC12IFG = 0;

  for (;;)
    {


      if (mmc_init () == MMC_SUCCESS)	// Karte gefunden
	card_state |= 1;
      else
	card_state = 0;		// keine Karte

      mmc_read_block (0x0, 512);	// wichtig: Die Blocklaenge muss gleich einer Zweierpotenz sein
      mmc_read_register (MMC_READ_CSD, 16);
      mmc_write_register (MMC_WRITE_CSD, 16);
      // Daten Schreiben: Erstmal an Adresse 0 die 0, an Adr. 1 die 1 usw. bis 512 (mit Bitmaske 0xff
      // u. cast auf unsigned char)
      mmc_write_block (0x00000000);

    }


  return;
}
