// mmc.c : MultiMediaCard-Funktionen: Lesen, Schreiben, Initialisieren ...
//
// Die MMC wird im unprotected SPI-Modus verwendet.
//
// Rolf Freitag 5/2003
// 



// MMC Lib 
#ifndef _MMCLIB_C
#define _MMCLIB_C
//---------------------------------------------------------------------
#include "mmc.h"

#include  "MSP430x14x.H"
#include  "math.h"
#include  "string.h"

char mmc_buffer[512] = { 0 };	// Buffer
unsigned int block_length = 512;	// Blocklaenge: Zum Schreiben sind immer 512 Byte noetig, z. Lesen 2^n

extern char card_state;		// Zustand der Karte: 0 fuer nicht gefunden, 1 fuer gefunden (init successfull)


//---------------------------------------------------------------------




// setup usart1 in spi mode
void
uart1_spi_mode_init (void)
{
  ME2 |= USPIE1;		// Enable USART1 SPI mode
  UTCTL1 = CKPH | SSEL1 | SSEL0 | STC;	// SMCLK, 3-pin mode, clock idle low, data valid on rising edge, UCLK delayed
  UCTL1 = CHAR | SYNC | MM;	// 8-bit SPI Master **SWRST**
  UBR01 = 0x02;			// 0x02: UCLK/2 (4 MHz), works also with 3 and 4
  UBR11 = 0x00;			// -"-
  UMCTL1 = 0x00;		// no modulation
  P5SEL |= 0x0E;		// P5.1-3 SPI option select
  P5DIR |= 0x01;		// P5.0 output direction
  P5OUT = 0xff;
  return;
}


// Initialisieren
char
mmc_init (void)
{
  char rValue = MMC_INIT_ERROR;
  unsigned int i = 0;

  // setup the SPI:
  // clock idle LOW 
  // data valid on rising edge 
  // setup the usart
  uart1_spi_mode_init ();

  // begin MMC initialization sequence 
  // card must be clocked at least 64 times
  // chip select mmc
  while (!(IFG2 & UTXIFG1));	// USART1 TX buffer ready (empty)?   
  EN_HIGH ();
  // Send 80 Clocks to start bus communication
  // send DUMMY (0xFF) 10 times 
  for (i = 0; i < 10; ++i)
    spi_write (DUMMY);
  while ((IFG2 & UTXIFG1) == 0);	// wait while not ready / for RX
  // Assert the chip select (CS) pin, (active low).
  EN_LOW ();

  // give the MMC the required clocks to finish up what ever it needs to do 
//  for (i = 0; i < 8; ++i)
//    spi_write (DUMMY);

  // command to put MMC into (unprotected) SPI mode (reset command, CMD0) 
  // command = 0x40 
  // all args = 0x00 
  // crc = 0x95 (pre-calculated) 
  mmc_send_command (0x40, 0x00, 0x00, 0x00, 0x00, 0x95);

  // Send 8 pulses of delay and wait for valid response.
  if (mmc_wait_for_response (0x01) == MMC_SUCCESS)
    {

/*
	  while (!(IFG2 & UTXIFG1));	// USART1 TX buffer ready (empty)?   
	  EN_LOW ();

	  // Send 8 Clock pulses of delay, CMD1, 8 pulses of delay and wait for valid response.
	  spi_write (DUMMY);

	  while (!(IFG2 & UTXIFG1));	// USART1 TX buffer ready (empty)?   
	  EN_HIGH ();
*/


      // Send 8 Clock pulses of delay, CMD1, 8 pulses of delay and wait for valid response.
      spi_write (DUMMY);
      i = 0xff;
      do			// MMC CMD1, Init
	mmc_send_command (0x41, 0x00, 0x00, 0x00, 0x00, 0xFF);
      while ((mmc_wait_for_response (MMC_R1_RESPONSE) != MMC_SUCCESS)
	     && (--i > 0));

      // Send 8 Clock pulses of delay.
      spi_write (DUMMY);

      // initialization fails if the card didn't come out of idle state 
      if (i != 0)
	rValue = MMC_SUCCESS;	// card came out of idle state  w / out timing out

    }
  // give the MMC the required clocks to finish up what ever it needs to do 
//  for (i = 0; i < 8; ++i)
//      spi_write (DUMMY);

  while (!(IFG2 & UTXIFG1));	// USART1 TX buffer ready (empty)? 
  EN_HIGH ();

  // Send 8 Clock pulses of delay.
  spi_write (DUMMY);

  if (rValue == MMC_SUCCESS)	// card found
    card_state |= 0x01;
  else
    card_state &= ~0x01;
  EN_HIGH ();
  return rValue;		// 0 if ok, 4 if init error
}				// init




//---------------------------------------------------------------------
// Wait for expected value (0x00, 0x01, 0xFE) or timeout after 255 reads (with CS low)
char
mmc_wait_for_response (const char expected_c)
{
  char rValue = MMC_SUCCESS;
  char val;
  unsigned char count = 0xFF;

  do
    val = spi_read ();
  while ((val != expected_c) && (--count > 0));

  if (count == 0)		// timed-out while waiting for a valid response 
    rValue = MMC_TIMEOUT_ERROR;	// 0xFF 

  return rValue;
}



char				// The card will respond with a standard response token followed by a data block suffixed with a 16 bit CRC.
mmc_read_block (const unsigned long int address, const unsigned int count)
{
  unsigned long i = 0;
  char rvalue = MMC_RESPONSE_ERROR;

  // Set the block length to read 
  if (mmc_set_block_length (count) == MMC_SUCCESS)	// block length could be set
    {

      // SS = LOW (on) 
      while (!(IFG2 & UTXIFG1));	// USART1 TX buffer ready (empty)? 
      EN_LOW ();
      // send read command 
      mmc_send_command (MMC_READ_SINGLE_BLOCK,
			HIGH (address >> 16),
			LOW (address >> 16), HIGH (address),
			LOW (address), 0xFF);
      // Send 8 Clock pulses of delay, check if the MMC acknowledged the read block command 
      // it will do this by sending an affirmative response 
      // in the R1 format (0x00 is no errors) 
      if (mmc_wait_for_response (MMC_R1_RESPONSE) == MMC_SUCCESS)
	{
	  // now look for the data token to signify the start of
	  // the data 
	  if (mmc_wait_for_response (MMC_START_DATA_BLOCK_TOKEN) ==
	      MMC_SUCCESS)
	    {
	      // clock the actual data transfer and receive the bytes; spi_read automatically finds the Data Block
	      for (i = 0; i < block_length; i++)
		mmc_buffer[i] = spi_read ();	// is executed with card inserted

	      // get CRC bytes (not really needed by us, but required by MMC)
	      spi_read ();
	      spi_read ();
	      rvalue = MMC_SUCCESS;
	    }
	  else
	    {
	      // the data token was never received 
	      rvalue = MMC_DATA_TOKEN_ERROR;	// 3
	    }
	}
      else
	{
	  // the MMC never acknowledge the read command 
	  rvalue = MMC_RESPONSE_ERROR;	// 2
	}
    }
  else
    {
      rvalue = MMC_BLOCK_SET_ERROR;	// 1
    }
  while (!(IFG2 & UTXIFG1));	// USART1 TX buffer ready (empty)? 
  EN_HIGH ();
  spi_write (DUMMY);
  return rvalue;
}				// mmc_read_block



//---------------------------------------------------------------------
char
mmc_write_block (const unsigned long int address)
{
  unsigned long i = 0;
  char rvalue = MMC_RESPONSE_ERROR;	// MMC_SUCCESS;      
  char c = 0x00;



  // Set the block length to read 
  if (mmc_set_block_length (512) == MMC_SUCCESS)	// block length could be set
    {
      // give the MMC the required clocks to finish up what ever it needs to do 
//      for (i = 0; i < 8; ++i)
//      spi_write (DUMMY);

      // SS = LOW (on) 
      while (!(IFG2 & UTXIFG1));	// USART1 TX buffer ready (empty)?       
      EN_LOW ();

      // give the MMC the required clocks to finish up what ever it needs to do 
//      for (i = 0; i < 8; ++i)
//      spi_write (DUMMY);

      // send write command 
      mmc_send_command (MMC_CMD_WRITEBLOCK,
			HIGH (address >> 16),
			LOW (address >> 16), HIGH (address),
			LOW (address), 0xFF);

      // check if the MMC acknowledged the write block command 
      // it will do this by sending an affirmative response 
      // in the R1 format (0x00 is no errors) 
      if (mmc_wait_for_response (MMC_R1_RESPONSE) == MMC_SUCCESS)
	{

//            for (i = 0; i < 8; ++i)
	  spi_write (DUMMY);

	  // send the data token to signify the start of the data 
	  spi_write (MMC_START_DATA_BLOCK_WRITE);
	  // clock the actual data transfer and transmitt the bytes 
	  for (i = 0; i < 512; i++)
	    spi_write ((char) (i & 0xff));	// mmc_buffer[i];       Test: i & 0xff
	  // put CRC bytes (not really needed by us, but required by MMC)
	  spi_write (DUMMY);
	  spi_write (DUMMY);
	  // read the data response xxx0<status>1 : status 010: Data accected, status 101: Data
	  //   rejected due to a crc error, status 110: Data rejected due to a Write error.
	  c = spi_read ();
	  c &= 0x1f;
	  switch (c)
	    {
	    case 0x05:
	      rvalue = MMC_SUCCESS;
	      break;
	    case 0x0b:
	      rvalue = MMC_CRC_ERROR;
	      break;
	    case 0x0d:
	      rvalue = MMC_WRITE_ERROR;
	      break;
	    default:
	      rvalue = MMC_OTHER_ERROR;
	      break;
	    }
	}
      else
	{
	  // the MMC never acknowledge the write command 
	  rvalue = MMC_RESPONSE_ERROR;	// 2
	}
    }
  else
    {
      rvalue = MMC_BLOCK_SET_ERROR;	// 1
    }
  // give the MMC the required clocks to finish up what ever it needs to do 
  for (i = 0; i < 9; ++i)
    spi_write (DUMMY);

  while (!(IFG2 & UTXIFG1));	// USART1 TX buffer ready (empty)? 
  EN_HIGH ();
  // Send 8 Clock pulses of delay.
  spi_write (DUMMY);
  return rvalue;
}				// mmc_write_block





//---------------------------------------------------------------------
void
mmc_send_command (const char cmd, const char arg1, const char arg2,
		  const char arg3, const char arg4, const char crc)
{

  // Send command ( EN_LOW has to be set before)
  spi_write (cmd);
  // Send arguments 
  spi_write (arg1);
  spi_write (arg2);
  spi_write (arg3);
  spi_write (arg4);
  // Send CRC 
  spi_write (crc);
  return;
}


//---------------  blocklength muss 2^n sein------------------------------------------------------
char				// ab 6.5. uerberfluessig, weil blocklaenge fest auf 512
mmc_set_block_length (const unsigned int blocklength)
{
  char rValue = MMC_TIMEOUT_ERROR;
  char i = 0;

  // SS = LOW (on) 
  while (!(IFG2 & UTXIFG1));	// USART1 TX buffer ready (empty)? 
  EN_LOW ();


  // Set the block length to read 
  mmc_send_command (MMC_SET_BLOCKLEN, 0, 0, HIGH (blocklength),
		    LOW (blocklength), 0xFF);
  // get response from MMC - make sure that its 0x00 (R1 ok response format)
  // Caution: if no card is inserted 0x00 (R1 ok response format) will be read!
  rValue = mmc_wait_for_response (MMC_R1_RESPONSE);

  // give the MMC the required clocks to finish up what ever it needs to do 
//   for (i = 0; i < 8; ++i)
//      spi_write (DUMMY);

  // CS = HIGH (off) 
  while (!(IFG2 & UTXIFG1));	// USART1 TX buffer ready (empty)? 
  EN_HIGH ();

  // Send 8 Clock pulses of delay.
  spi_write (DUMMY);
  if (rValue == MMC_SUCCESS)
    block_length = blocklength;
  else
    block_length = 0;
  return rValue;
  EN_HIGH ();
  return rValue;
}				// block_length


void
spi_write (const char x)
{
//  while (!(UTCTL1 & TEXPT));  // wait for TX Empty      ggf. define TEXPT 0x01
  while (!(IFG2 & UTXIFG1));	// wait while not ready / for RX
  TXBUF1 = x;			// write
}



char
spi_read (void)
{
  spi_write (0xff);
//  while (!(IFG2 & UTXIFG1));  // USART1 TX buffer ready (empty)?  
  while (!(IFG2 & URXIFG1));	// wait for RX 
  return RXBUF1;		// return buffer
}




char				// Reading the contents of the CSD and CID registers in SPI mode is a simple read-block transaction.
mmc_read_register (const char cmd_register, const unsigned char length)
{
  char uc = 0;
  char rvalue = MMC_TIMEOUT_ERROR;
//  char i = 0;

  if (mmc_set_block_length (length) == MMC_SUCCESS)
    {
      while (!(IFG2 & UTXIFG1));	// USART1 TX buffer ready (empty)? 
      EN_LOW ();
      // CRC not used: 0xff as last byte
      mmc_send_command (cmd_register, 0, 0, 0, 0, 0xff);


      // wait for response 
      // in the R1 format (0x00 is no errors) 
      if (mmc_wait_for_response (MMC_R1_RESPONSE) == MMC_SUCCESS)
	{

	  rvalue = mmc_wait_for_response (MMC_START_DATA_BLOCK_TOKEN);
	  if (rvalue == MMC_SUCCESS)
	    for (uc = 0; uc < length; uc++)
	      mmc_buffer[uc] = spi_read ();

	  // give the MMC the required clocks to finish up what ever it needs to do 
//  for (i = 0; i < 8; ++i)
//    spi_write (DUMMY);

	  // get CRC bytes (not really needed by us, but required by MMC)
	  spi_read ();
	  spi_read ();
	}
      else
	rvalue = MMC_RESPONSE_ERROR;
      // CS = HIGH (off) 
      while (!(IFG2 & UTXIFG1));	// USART1 TX buffer ready (empty)? 
      EN_HIGH ();

      // Send 8 Clock pulses of delay.
      spi_write (DUMMY);
    }
  EN_HIGH ();
  return rvalue;
}				// mmc_read_register






char
mmc_write_register (const char cmd_register, const unsigned char length)
{
  unsigned char uc = 0;
  char rvalue = MMC_TIMEOUT_ERROR;
  char i = 0;
  char c = 0x00;

  // give the MMC the required clocks to finish up what ever it needs to do 
//  for (i = 0; i < 8; ++i)
//    spi_write (DUMMY);
  if (mmc_set_block_length (length) == MMC_SUCCESS)
    {
      while (!(IFG2 & UTXIFG1));	// USART1 TX buffer ready (empty)? 
      EN_LOW ();
      mmc_send_command (cmd_register, 0, 0, 0, 0, 0xFF);
      rvalue = mmc_wait_for_response (MMC_R1_RESPONSE);	// Response R1 == 0x00: no error

      for (i = 0; i < 8; ++i)
	spi_write (DUMMY);

      if (rvalue == MMC_SUCCESS)
	{
	  spi_write (MMC_START_DATA_BLOCK_TOKEN);
	  for (uc = 0; uc < length; uc++)
	    spi_write (mmc_buffer[uc]);
	}

      rvalue = mmc_wait_for_response (MMC_R1_RESPONSE);	// Response R1 == 0x00: no error

      // give the MMC the required clocks to finish up what ever it needs to do 
      for (i = 0; i < 9; ++i)
	spi_write (DUMMY);
      // put CRC bytes (not really needed by us, but required by MMC)
      spi_write (DUMMY);
      spi_write (DUMMY);
      // read the data response xxx0<status>1 : status 010: Data accected, status 101: Data
      //   rejected due to a crc error, status 110: Data rejected due to a Write error.
      c = spi_read ();
      c &= 0x1f;
      switch (c)
	{
	case 0x05:
	  rvalue = MMC_SUCCESS;
	  break;
	case 0x0b:
	  rvalue = MMC_CRC_ERROR;
	  break;
	case 0x0d:
	  rvalue = MMC_WRITE_ERROR;
	  break;
	default:
	  rvalue = MMC_OTHER_ERROR;
	  break;
	}

      // give the MMC the required clocks to finish up what ever it needs to do 
      for (i = 0; i < 9; ++i)
	spi_write (DUMMY);

      // CS = HIGH (off) 
      while (!(IFG2 & UTXIFG1));	// USART1 TX buffer ready (empty)? 
      EN_HIGH ();

      // give the MMC the required clocks to finish up what ever it needs to do 
//  for (i = 0; i < 8; ++i)
      spi_write (DUMMY);
    }
  EN_HIGH ();
  return rvalue;
}				// mmc_write_register



//---------------------------------------------------------------------
#endif /* _MMCLIB_C */
